"""Utils for jax."""

import jax
import jax.numpy as jnp
from jax._src.typing import Array  # TODO: Check why Array is not in jax.typing like in the docs
from jax.typing import ArrayLike


# %% Tree utils
def tree_stack(trees, axis=0):
    """Stack pytrees along a given axis."""
    return jax.tree_map(lambda *v: jnp.stack(v, axis), *trees)


def tree_unstack(tree):
    leaves, treedef = jax.tree_flatten(tree)
    return [treedef.unflatten(leaf) for leaf in zip(*leaves, strict=True)]


def tree_pad(traj, pad_width, pad_mode="constant", pad_constant_values=0):
    return jax.tree_map(
        lambda x: jnp.pad(
            x, pad_width[: x.ndim], mode=pad_mode, constant_values=pad_constant_values
        ),
        traj,
    )


def compute_smart_bins(
    dataset: list[dict[str, ArrayLike]],
    nb_bins: int,
) -> list[list[dict[str, Array]]]:
    """
    Group trajectories in bins of similar length to limit the padding required.

    Thanks to Mathis Hardion for finding this elegant solution.
    """
    if nb_bins == 1:
        return [dataset]
    nb_sequences = len(dataset)
    sorted_dataset = sorted(dataset, key=lambda x: len(x["observations"]), reverse=True)
    len_deltas = [
        len(sorted_dataset[i]["observations"]) - len(sorted_dataset[i + 1]["observations"])
        for i in range(nb_sequences - 1)
    ]
    len_deltas = jnp.array(len_deltas)
    top_k_delta_indices = jnp.argpartition(len_deltas, -nb_bins + 1)[-nb_bins + 1 :]
    top_k_delta_indices = jnp.sort(top_k_delta_indices)

    bins = []
    start_index = 0
    for end_index in top_k_delta_indices:
        bins.append(sorted_dataset[start_index : end_index + 1])
        start_index = end_index + 1
    bins.append(sorted_dataset[start_index:])

    return bins


def smart_pad_dataset(
    dataset: list[dict[str, Array]],
    nb_bins: int,
    pad_mode: str = "constant",
    pad_constant_values: int = 0,
) -> list[dict[str, Array]]:
    """
    Pad dataset with smart bins.

    Output is a dataset whose trajectories are padded with nb_bins different lengths
    such that the total padding is minimized.
    """
    bins = compute_smart_bins(dataset, nb_bins)
    padded_dataset = []
    for traj_bin in bins:
        max_traj_len = max(len(traj["observations"]) for traj in traj_bin)
        for traj in traj_bin:
            traj_len = len(traj["observations"])
            pad_width = ((0, max_traj_len - traj_len), (0, 0), (0, 0), (0, 0))
            traj = tree_pad(traj, pad_width, pad_mode, pad_constant_values)
            traj["mask"] = jnp.concatenate([jnp.ones(traj_len), jnp.zeros(max_traj_len - traj_len)])
            padded_dataset.append(traj)

    return padded_dataset



"""Core module."""

from collections.abc import Callable
from typing import Any


def make_yamlizable[T](obj: T, /, replace_fn: Callable[[Any], Any] | None = None) -> T:
    """
    Recursively convert non-serializable objects to their names in a copy of a dictionary.

    Args:
        obj: The dictionary to make YAML-serializable.
        replace_fn: A function to replace non-serializable objects. If None,
            replace callable objects with their names.

    Returns:
        A YAML-serializable dictionary.
    """
    # By default, replace callable objects with their names
    if replace_fn is None:
        replace_fn = lambda v: v.__name__ if callable(v) else v

    if isinstance(obj, dict):
        return {k: make_yamlizable(v, replace_fn) for k, v in obj.items()}
    if isinstance(obj, list):
        return [make_yamlizable(item, replace_fn) for item in obj]
    if isinstance(obj, tuple):
        return tuple(make_yamlizable(item, replace_fn) for item in obj)
    if isinstance(obj, float) and obj.is_integer():
        return int(obj)
    return replace_fn(obj)



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from kajihs_utils.numpy_utils import IncompatibleShapeError, find_closest


# TODO: Check and fix those tests
class TestFindClosest:
    def test_scalar_inputs_single_target(self):
        # x: (N,), target: scalar
        x = np.array([0, 10, 20, 30])
        target = 12
        # The closest to 12 is 10 at index 1
        result = find_closest(x, target)
        assert result == 1

    def test_scalar_inputs_multiple_targets(self):
        # x: (N,), targets: (M,)
        x = np.array([0, 10, 20, 30])
        targets = np.array([2, 26])
        # Closest to 2 is 0 at index 0, closest to 26 is 30 at index 3
        result = find_closest(x, targets)
        assert_array_equal(result, np.array([0, 3]))

    def test_vector_inputs_single_target(self):
        # x: (N, D), target: (D,)
        x = np.array([[0, 0], [10, 10], [20, 20]])
        target = np.array([6, 5])
        # Distances:
        # to [0,0] ~ sqrt(6^2+5^2)=sqrt(61)
        # to [10,10] ~ sqrt((10-6)^2+(10-5)^2)=sqrt(16+25)=sqrt(41)=closest
        # to [20,20] ~ sqrt((20-6)^2+(20-5)^2) large number
        result = find_closest(x, target)
        assert result == 1

    def test_vector_inputs_multiple_targets(self):
        # x: (N, D), targets: (M, D)
        x = np.array([[0, 0], [10, 10], [20, 20]])
        targets = np.array([[-1, -1], [15, 12]])
        # Distances to [-1,-1]:
        #   [0,0]: sqrt(1+1)=sqrt(2)
        #   [10,10]: large
        #   [20,20]: larger
        # Closest: index 0
        # Distances to [15,12]:
        #   [0,0]: sqrt(15^2+12^2)=sqrt(225+144)=sqrt(369)
        #   [10,10]: sqrt(5^2+2^2)=sqrt(29)=closest
        #   [20,20]: sqrt((20-15)^2+(20-12)^2)=sqrt(25+64)=sqrt(89)
        # Closest: index 1
        result = find_closest(x, targets)
        assert_array_equal(result, np.array([0, 1]))

    def test_matrix_inputs_single_target(self):
        # x: (N, H, W), target: (H, W)
        x = np.array([[[0, 0], [0, 0]], [[10, 10], [10, 10]], [[20, 20], [20, 20]]])
        target = np.array([[2, 2], [2, 2]])
        # Distances (fro norm):
        # to [[[0,0],[0,0]]] = sqrt(2^2+2^2+2^2+2^2)=sqrt(16)=4
        # to [[[10,10],[10,10]]] = large
        # to [[[20,20],[20,20]]] = larger
        # Closest: index 0
        result = find_closest(x, target, norm_ord="fro")
        assert result == 0

    def test_matrix_inputs_multiple_targets(self):
        # x: (N, H, W)
        x = np.array([[[0, 0], [0, 0]], [[10, 10], [10, 10]], [[20, 20], [20, 20]]])
        targets = np.array([
            [[0, 0], [1, 1]],  # close to first
            [[19, 19], [19, 19]],  # close to last
        ])
        # First target distances:
        #   to x[0]: fro sqrt(0^2+0^2+(1^2)+(1^2)=sqrt(2)=1.414..
        #   to x[1]: fro sqrt((10^2+10^2)+(9^2+9^2)) large
        #   to x[2]: even larger
        # Closest: index 0
        # Second target distances:
        #   to x[0]: big
        #   to x[1]: sqrt((9^2 four times)= sqrt(81*4)=sqrt(324)=18
        #   to x[2]: sqrt((1^2 four times)= sqrt(4)=2 closest
        # Closest: index 2
        result = find_closest(x, targets, norm_ord="fro")
        assert_array_equal(result, np.array([0, 2]))

    def test_four_dimensional_inputs(self):
        # x: (N, A, B, C) e.g. (3,2,2,2)
        x = np.array([
            np.zeros((2, 2, 2)),
            np.ones((2, 2, 2)) * 10,
            np.ones((2, 2, 2)) * 20,
        ])
        target = np.ones((2, 2, 2)) * 9
        # Flattened norms:
        # Dist to x[0]: sqrt(9^2 * 8 elements)= sqrt(81*8)= sqrt(648)
        # Dist to x[1]: sqrt((10-9)^2 * 8)= sqrt(1*8)= sqrt(8)=2.828...
        # Dist to x[2]: sqrt((20-9)^2 * 8)= large
        # Closest: index 1
        result = find_closest(x, target)
        assert result == 1

    def test_multiple_targets_for_four_dimensional(self):
        x = np.array([
            np.zeros((2, 2, 2)),
            np.ones((2, 2, 2)) * 10,
            np.ones((2, 2, 2)) * 20,
        ])
        targets = np.array([
            np.ones((2, 2, 2)),  # closer to 0 (distance large) or 10?
            np.ones((2, 2, 2)) * 15,  # closer to 10 or 20?
        ])
        # For target ~1:
        # Dist to x[0]: sqrt((1-0)^2 *8)= sqrt(1*8)= sqrt(8)
        # Dist to x[1]: sqrt((10-1)^2 *8)= sqrt(81*8)=sqrt(648)
        # Dist to x[2]: sqrt((20-1)^2 *8)= even bigger
        # Closest: index 0
        # For target ~15:
        # Dist to x[0]: sqrt(15^2*8) large
        # Dist to x[1]: sqrt((10-15)^2*8)= sqrt(25*8)= sqrt(200)
        # Dist to x[2]: sqrt((20-15)^2*8)= sqrt(25*8)= sqrt(200)
        # Ties between index 1 and 2, np.argmin returns the first min, so index 1.
        result = find_closest(x, targets)
        assert_array_equal(result, np.array([0, 1]))

    def test_single_target_same_shape_returns_int(self):
        x = np.array([[0, 1], [2, 3], [4, 5]])
        target = np.array([3, 4])
        # This should return a single integer because target is the same shape as an element
        result = find_closest(x, target)
        assert isinstance(result, np.integer)

    def test_multiple_targets_return_array(self):
        x = np.array([[0, 1], [2, 3], [4, 5]])
        targets = np.array([[1, 1], [3, 3]])
        # This should return an array because we have multiple targets
        result = find_closest(x, targets)
        assert isinstance(result, np.ndarray)
        assert result.shape == (2,)

    def test_incompatible_shape_error_single_target(self):
        x = np.array([[0, 10, 20], [30, 40, 50]])  # shape (4,)
        target = np.array([1, 2])  # shape (2,) incompatible with (N,) elements
        with pytest.raises(IncompatibleShapeError):
            find_closest(x, target)

    def test_incompatible_shape_error_multiple_targets(self):
        x = np.array([[0, 0], [1, 1], [2, 2]])  # shape (3,2)
        targets = np.array([[0], [1]])  # shape (2,1), incompatible with (2,)
        with pytest.raises(IncompatibleShapeError):
            find_closest(x, targets)

    def test_norm_ord_none(self):
        # Check that passing norm_ord=None is acceptable and uses default norm (2-norm)
        x = np.array([[0, 0], [10, 10], [20, 20]])
        target = np.array([6, 5])
        result_default = find_closest(x, target, norm_ord=None)
        result_2 = find_closest(x, target, norm_ord=2)
        assert result_default == result_2

    def test_invalid_norm_ord_for_non_matrix_with_fro(self):
        # "fro" norm is defined for matrices (2D), but let's see if it raises an error for vectors.
        x = np.array([0, 10, 20, 30])  # shape (4,), scalar dimension
        target = 12
        # numpy.linalg.norm with 'fro' and 1D array works but treats it as a 2D array with shape (4,1)
        # which leads to a ValueError because 'fro' is only defined for 2D matrices.
        # We can test if this raises a ValueError.
        with pytest.raises(ValueError):
            find_closest(x, target, norm_ord="fro")

    def test_invalid_norm_ord_for_non_matrix_with_nuc(self):
        # "nuc" norm (nuclear norm) is only defined for 2D arrays.
        # Here we test with a 1D array and expect a ValueError.
        x = np.array([0, 10, 20, 30])
        target = 12
        with pytest.raises(ValueError):
            find_closest(x, target, norm_ord="nuc")

    def test_nuc_norm_for_2d_arrays(self):
        x2 = np.array([[[0, 0]], [[10, 10]], [[20, 20]]])  # shape (3, 1, 2)
        target2 = np.array([[6, 5]])  # shape (1,2)
        result = find_closest(x2, target2, norm_ord="nuc")
        # Should return a single index
        assert result == 1

    def test_custom_norm_ord_int(self):
        # Test a custom integer norm (like 1-norm)
        x = np.array([[0, 0], [10, 10], [20, 20]])
        target = np.array([6, 5])
        # With 1-norm:
        # Distances:
        # to [0,0]: |6|+|5|=11
        # to [10,10]: |10-6|+|10-5|=4+5=9 closest
        # to [20,20]: big
        result = find_closest(x, target, norm_ord=1)
        assert result == 1

    def test_multiple_targets_and_int_norm(self):
        x = np.array([[0, 0], [10, 10], [20, 20]])
        targets = np.array([[1, 1], [15, 15]])
        # With 1-norm:
        # For [1,1]:
        #   to [0,0]: sum(|1|+|1|)=2
        #   to [10,10]: sum(|9|+|9|)=18
        #   to [20,20]: sum(|19|+|19|)=38
        # closest: 0
        # For [15,15]:
        #   to [0,0]: sum(|15|+|15|)=30
        #   to [10,10]: sum(|5|+|5|)=10
        #   to [20,20]: sum(|5|+|5|)=10 tie, argmin picks first: index 1
        result = find_closest(x, targets, norm_ord=1)
        assert_array_equal(result, [0, 1])



"""
Utils for arithmetic.

close_factors and almost_factors taken from:
https://code.visualstudio.com/api/language-extensions/semantic-highlight-guide
"""


def closest_factors(n: int, /) -> tuple[int, int]:
    """
    Find the closest pair of factors.

    Args:
        n: The number to find factors for.

    Returns:
        A tuple containing the two closest factors of n, the larger first.

    Example:
        >>> closest_factors(99)
        (11, 9)
    """
    factor1 = 0
    factor2 = n
    while factor1 + 1 <= factor2:
        factor1 += 1
        if n % factor1 == 0:
            factor2 = n // factor1

    return factor1, factor2


def almost_factors(n: int, /, ratio: float = 0.5) -> tuple[int, int]:
    """
    Find a pair of factors that are close enough.

    Args:
        n: The number to almost-factorize.
        ratio: The threshold ratio between both factors.

    Returns:
        A tuple containing the first two numbers factoring to n or more such
        that factor 1 is at most 1/ratio times larger than factor 2.

    Example:
        >>> almost_factors(10, ratio=0.5)
        (4, 3)
    """
    while True:
        factor1, factor2 = closest_factors(n)
        if ratio * factor1 <= factor2:
            break
        n += 1
    return factor1, factor2



"""Totally typed, plausibly practical, and remarkably random utilities—for me, and maybe for you too."""

from kajihs_utils.core import bisect_predicate, get_first, is_sorted

__all__ = ["bisect_predicate", "get_first", "is_sorted"]



"""Useful protocols for structural subtyping."""

from typing import (
    Any,
    Protocol,
    TypeVar,
    runtime_checkable,
)

_T_contra = TypeVar("_T_contra", contravariant=True)

# Comparison protocols


@runtime_checkable
class SupportsDunderLT(Protocol[_T_contra]):  # noqa: D101
    def __lt__(self, other: _T_contra, /) -> bool: ...


@runtime_checkable
class SupportsDunderGT(Protocol[_T_contra]):  # noqa: D101
    def __gt__(self, other: _T_contra, /) -> bool: ...


@runtime_checkable
class SupportsDunderLE(Protocol[_T_contra]):  # noqa: D101
    def __le__(self, other: _T_contra, /) -> bool: ...


@runtime_checkable
class SupportsDunderGE(Protocol[_T_contra]):  # noqa: D101
    def __ge__(self, other: _T_contra, /) -> bool: ...


@runtime_checkable
class SupportsAllComparisons[T](  # noqa: D101
    SupportsDunderLT[T],
    SupportsDunderGT[T],
    SupportsDunderLE[T],
    SupportsDunderGE[T],
    Protocol,
): ...


type SupportsRichComparison = SupportsDunderLT[Any] | SupportsDunderGT[Any]
SupportsRichComparisonT = TypeVar("SupportsRichComparisonT", bound=SupportsRichComparison)



"""Tools for numpy."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Self, override

import numpy as np
from numpy import dtype, float64, floating, int_, ndarray

if TYPE_CHECKING:
    from collections.abc import Iterable

    from numpy.typing import ArrayLike, NDArray

type Norm = float | Literal["fro", "nuc"]
type AnyFloat = float | float64


class IncompatibleShapeError(ValueError):
    """Shapes of input arrays are incompatible for a given function."""

    def __init__(self, arr1: NDArray[Any], arr2: NDArray[Any], obj: Any) -> None:
        super().__init__(
            f"Shapes of inputs arrays {arr1.shape} and {arr2.shape} are incompatible for {obj.__name__}"
        )


# TODO: Add axis parameters
def find_closest[T](
    x: Iterable[T] | ArrayLike,
    targets: Iterable[T] | T | ArrayLike,
    norm_ord: Norm | None = None,
) -> ndarray[tuple[int], dtype[int_]] | int_:
    """
    Find the index of the closest element(s) from `x` for each target in `targets`.

    Given one or multiple `targets` (vectors vectors or scalars),
    this function computes the distance to each element in `x` and returns the
    indices of the closest matches. If `targets` is of the same shape as an
    element of `x`, the function returns a single integer index. If `targets`
    contains multiple elements, it returns an array of indices corresponding to
    each target.

    If the dimensionality of the vectors in `x` is greater than 2, the vectors
    will be flattened into 1D before computing distances.

    Args:
        x: An iterable or array-like collection of elements (scalars, vectors,
            or higher-dimensional arrays). For example, `x` could be an array of
            shape `(N,)` (scalars), `(N, D)` (D-dimensional vectors),
            `(N, H, W)` (2D arrays), or higher-dimensional arrays.
        targets: One or multiple target elements for which you want to find the
            closest match in `x`. Can be a single scalar/vector/array or an
            iterable of them.
            Must be shape-compatible with the elements of `x`.
        norm_ord: The order of the norm used for distance computation.
            Uses the same conventions as `numpy.linalg.norm`.

    Returns:
        An array of indices. If a single target was given, a single index is
        returned. If multiple targets were given, an array of shape `(M,)` is
        returned, where `M` is the number of target elements. Each value is the
        index of the closest element in `x` to the corresponding target.

    Raises:
        IncompatibleShapeError: If `targets` cannot be broadcast or reshaped to
            match the shape structure of the elements in `x`.

    Examples:
        >>> import numpy as np
        >>> x = np.array([0, 10, 20, 30])
        >>> int(find_closest(x, 12))
        1
        >>> # Multiple targets
        >>> find_closest(x, [2, 26])
        array([0, 3])

        >>> # Using vectors
        >>> x = np.array([[0, 0], [10, 10], [20, 20]])
        >>> int(find_closest(x, [6, 5]))  # Single target vector
        1
        >>> find_closest(x, [[-1, -1], [15, 12]])  # Multiple target vectors
        array([0, 1])

        >>> # Higher dimensional arrays
        >>> x = np.array([[[0, 0], [0, 0]], [[10, 10], [10, 10]], [[20, 20], [20, 20]]])
        >>> int(find_closest(x, [[2, 2], [2, 2]]))
        0
        >>> find_closest(x, [[[0, 0], [1, 1]], [[19, 19], [19, 19]]])
        array([0, 2])
    """
    x = np.array(x)  # (N, vector_shape)
    targets = np.array(targets)
    vector_shape = x.shape[1:]

    # Check that shapes are compatible
    do_unsqueeze = False
    if targets.shape == vector_shape:
        targets = np.atleast_1d(targets)[np.newaxis, :]  # (M, vector_shape)
        do_unsqueeze = True
    elif targets.shape[1:] != vector_shape:
        raise IncompatibleShapeError(x, targets, find_closest)

    nb_vectors = x.shape[0]  # N
    nb_targets = targets.shape[0]  # M

    diffs = x[:, np.newaxis] - targets

    match vector_shape:
        case ():
            distances = np.linalg.norm(diffs[:, np.newaxis], ord=norm_ord, axis=1)
        case (_,):
            distances = np.linalg.norm(diffs, ord=norm_ord, axis=2)
        case (_, _):
            distances = np.linalg.norm(diffs, ord=norm_ord, axis=(2, 3))
        case _:  # Tensors
            # Reshape to 1d vectors
            diffs = diffs.reshape(nb_vectors, nb_targets, -1)
            distances = np.linalg.norm(diffs, ord=norm_ord, axis=2)

    closest_indices = np.argmin(distances, axis=0)
    if do_unsqueeze:
        closest_indices = closest_indices[0]

    return closest_indices


class Vec2d(ndarray[Literal[2], dtype[float64]]):
    """A 2D vector subclassing numpy.ndarray with .x and .y properties."""

    def __new__(cls, x: AnyFloat, y: AnyFloat) -> Self:  # noqa: D102
        obj = np.asarray([x, y], dtype=np.float64).view(cls)
        return obj

    @property
    def x(self) -> float:
        """X coordinate."""
        return self[0]

    @x.setter
    def x(self, value: float) -> None:
        self[0] = value

    @property
    def y(self) -> float:
        """Y coordinate."""
        return self[1]

    @y.setter
    def y(self, value: float) -> None:
        self[1] = value

    def magnitude(self) -> floating[Any]:
        """Magnitude or norm of the vector."""
        return np.linalg.norm(self)

    def normalized(self) -> Vec2d:
        """Return a normalized version of the vector."""
        mag = self.magnitude()
        return self if mag == 0 else Vec2d(self.x / mag, self.y / mag)

    def angle(self) -> float:
        """Return the angle (in degrees) between the vector and the positive x-axis."""
        return np.degrees(np.arctan2(self.y, self.x))

    def rotate(self, degrees_angle: float, center: tuple[float, float] = (0, 0)) -> Vec2d:
        """Rotates the vector counterclockwise by a given angle (in degrees) around the point (cx, cy)."""
        cx, cy = center[0], center[1]
        # Step 1: Translate the vector to the origin (subtract the center of rotation)
        translated_x = self.x - cx
        translated_y = self.y - cy

        # Step 2: Rotate the translated vector
        rad = np.radians(degrees_angle)
        rot_matrix = np.array([[np.cos(rad), -np.sin(rad)], [np.sin(rad), np.cos(rad)]])
        rotated_vector = rot_matrix @ np.array([translated_x, translated_y])

        # Step 3: Translate the vector back to its original position
        new_x = rotated_vector[0] + cx
        new_y = rotated_vector[1] + cy

        return Vec2d(new_x, new_y)

    @override
    def __repr__(self) -> str:
        return f"Vec2d({self.x:.2f}, {self.y:.2f})"



from math import isclose

from kajihs_utils.numpy_utils import Vec2d

ABS_TOL = 1e-12


def test_vec2d_creation():
    v = Vec2d(3.0, 4.0)
    assert isinstance(v, Vec2d)
    assert isclose(v.x, 3.0)
    assert isclose(v.y, 4.0)


def test_x_property():
    v = Vec2d(3.0, 4.0)
    assert isclose(v.x, 3.0)
    v.x = 5.0
    assert isclose(v.x, 5.0)


def test_y_property():
    v = Vec2d(3.0, 4.0)
    assert isclose(v.y, 4.0)
    v.y = 6.0
    assert isclose(v.y, 6.0)


def test_magnitude():
    v = Vec2d(3.0, 4.0)
    assert isclose(v.magnitude(), 5.0)  # Magnitude of (3, 4) is 5
    v_zero = Vec2d(0.0, 0.0)
    assert isclose(v_zero.magnitude(), 0.0)


def test_normalized():
    v = Vec2d(3.0, 4.0)
    normalized_v = v.normalized()
    assert isclose(normalized_v.magnitude(), 1.0)
    assert isclose(normalized_v.x, 3 / 5)
    assert isclose(normalized_v.y, 4 / 5)

    v_zero = Vec2d(0.0, 0.0)
    normalized_v_zero = v_zero.normalized()
    assert isclose(normalized_v_zero.x, 0.0)
    assert isclose(normalized_v_zero.y, 0.0)


def test_angle():
    v = Vec2d(1.0, 0.0)
    assert isclose(v.angle(), 0.0)  # Angle of (1, 0) is 0 degrees
    v2 = Vec2d(0.0, 1.0)
    assert isclose(v2.angle(), 90.0)  # Angle of (0, 1) is 90 degrees
    v3 = Vec2d(-1.0, 0.0)
    assert isclose(v3.angle(), 180.0)  # Angle of (-1, 0) is 180 degrees
    v4 = Vec2d(0.0, -1.0)
    assert isclose(v4.angle(), -90.0)  # Angle of (0, -1) is -90 degrees


def test_rotate():
    # Vector to rotate
    v = Vec2d(1.0, 0.0)

    # Rotate around the origin (0, 0), expected result after 90 degrees counterclockwise: (0, 1)
    rotated_v = v.rotate(90.0)
    assert isclose(rotated_v.x, 0.0, abs_tol=ABS_TOL)
    assert isclose(rotated_v.y, 1.0)

    # Rotate around the origin (0, 0), expected result after 180 degrees: (-1, 0)
    rotated_v_180 = v.rotate(180.0)
    assert isclose(rotated_v_180.x, -1.0)
    assert isclose(rotated_v_180.y, 0.0, abs_tol=ABS_TOL)

    # Rotate around a new center (cx=1, cy=1), expected result after 90 degrees counterclockwise
    # First, translate the vector to origin (1, 1), rotate, then translate back to (1, 1)
    v = Vec2d(2.0, 0.0)
    rotated_v_center = v.rotate(90.0, (1.0, 1.0))  # Rotate around (1, 1)
    assert isclose(rotated_v_center.x, 2.0)
    assert isclose(rotated_v_center.y, 2.0)

    # Rotate around the new center (cx=1, cy=1), expected result after 180 degrees: (0, 1)
    rotated_v_center_180 = v.rotate(180.0, (1.0, 1.0))
    assert isclose(rotated_v_center_180.x, 0.0, abs_tol=ABS_TOL)
    assert isclose(rotated_v_center_180.y, 2.0)


def test_repr():
    v = Vec2d(1.234, 5.678)
    assert repr(v) == "Vec2d(1.23, 5.68)"  # Test that it formats to 2 decimal places



"""kajihs-utils metadata."""

__module_name__ = "kajihs_utils"  # Not dynamic
__version__ = "0.10.1"

__authors__ = ["Julian Paquerot"]
__author_emails__ = ["itskajih@gmail.com"]

# === URLs ===
# https://packaging.python.org/en/latest/specifications/well-known-project-urls/#well-known-labels
__repo_url__ = "https://github.com/Kajiih/kajihs-utils"
__homepage_url__ = "https://github.com/Kajiih/kajihs-utils"
__issues_url__ = f"{__repo_url__}/issues"
__changelog_url__ = f"{__repo_url__}/blob/main/CHANGELOG.md"
__documentation_url__ = "https://Kajiih.github.io/kajihs-utils/"



"""Hatch build hook to bring dynamic metadata."""

import importlib.util
import sys
from pathlib import Path
from typing import Any, override

from hatchling.metadata.plugin.interface import MetadataHookInterface


class AboutMetadataHookOld(MetadataHookInterface):
    """Hatchling metadata hook that loads dynamic metadata from the `__about__.py` file."""

    @override
    def update(self, metadata: dict[str, Any]) -> None:
        """
        Update the metadata dictionary with values from the `__about__.py` file.

        Args:
            metadata: The dictionary containing the project metadata.

        Raises:
            ImportError: Metadata loading failed
        """
        # Dynamically load __about__.py
        about_path = Path(self.root) / "src" / "kajihs_utils" / "__about__.py"
        spec = importlib.util.spec_from_file_location("__about__", about_path)
        if spec is None or spec.loader is None:
            msg = f"Could not load metadata from {about_path}"
            raise ImportError(msg)

        about = importlib.util.module_from_spec(spec)
        sys.modules["__about__"] = about
        spec.loader.exec_module(about)

        # Map __about__ attributes to metadata fields
        authors: list[str] = about.__authors__
        emails: list[str] = about.__author_emails__
        metadata["authors"] = [
            {"name": name, "email": email} for name, email in zip(authors, emails, strict=True)
        ]
        metadata["urls"] = {
            "Repository": about.__repo_url__,
            "Homepage": about.__homepage_url__,
            "Issues": about.__issues_url__,
            "Changelog": about.__changelog_url__,
            "Documentation": about.__documentation_url__,
        }



"""Utils for better date and times, specifically using whenever."""

from datetime import datetime

from whenever import Instant, OffsetDateTime, PlainDateTime, SystemDateTime, ZonedDateTime

type ExactDateTime = ZonedDateTime | Instant | OffsetDateTime | SystemDateTime

type AllDateTime = PlainDateTime | ExactDateTime


def dt_to_system_datetime(dt: datetime) -> SystemDateTime:
    """Convert into exact time by assuming system timezone if necessary."""
    return (
        PlainDateTime.from_py_datetime(dt).assume_system_tz()
        if dt.tzinfo is None
        else SystemDateTime.from_py_datetime(dt)
    )



"""Utils for logging, specifically using Loguru."""

import inspect
import logging
import sys
from pathlib import Path
from typing import override

from loguru import logger
from rich.prompt import Prompt


def prompt(prompt: str, /) -> str:
    """Wrap rich.Prompt.ask to add newline and color."""
    val = Prompt.ask(f"\n[cyan bold]{prompt}[/cyan bold]")
    logger.debug(f'Prompt: "{prompt}" -> "{val}"')
    return val


class InterceptHandler(logging.Handler):
    """
    Intercepts logs from the standard logging module and forwards them to Loguru.

    Snippet from https://github.com/Delgan/loguru/tree/master
    """

    @override
    def emit(self, record: logging.LogRecord) -> None:
        """Forward log records from the standard logging system to Loguru."""
        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame = inspect.currentframe()
        depth = 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def setup_logging(prefix: str = "app", log_dir: str | Path = "logs") -> None:
    """
    Set up beautiful loguru logging in files and console.

    Redirects logging with Loguru, creates 2 logging files with and without
    colors and log to console.

    Args:
        prefix: Prefix for the log files without extensions.
        log_dir: Directory path to store log files.

    """
    # Redirect logging with Loguru
    logging.basicConfig(handlers=[InterceptHandler()], level=logging.WARNING, force=True)

    logger.remove()

    log_dir = Path(log_dir)
    logger.add(
        log_dir / f"{prefix}.log",
        level="DEBUG",
        format="{time:YYYY-MM-DD HH:mm:ss} | {level} | {message}",
        rotation="1 week",
        compression="zip",
    )
    logger.add(
        log_dir / f"{prefix}.clog",
        level="DEBUG",
        # format="{time:YYYY-MM-DD HH:mm:ss} | {level} | {message}",
        rotation="1 week",
        compression="zip",
        colorize=True,
    )
    logger.add(
        sys.stdout,
        level="INFO",
        # format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{message}</level>",
        format="<level>{message}</level>",
    )



"""Custom version provider for commitizen."""

import re
from pathlib import Path
from typing import final, override

from commitizen.config.base_config import BaseConfig
from commitizen.providers.base_provider import VersionProvider


@final
class AboutPyProvider(VersionProvider):
    """Get version from __about__.py, in the variable __version__."""

    def __init__(self, config: BaseConfig):
        """Initialize provider with the appropriate attributes."""
        super().__init__(config)
        del config
        self.file_path = Path("src") / "kajihs_utils" / "__about__.py"
        self.version_regex = re.compile(r'^__version__\s*=\s*[\'"]([^\'"]+)[\'"]', re.MULTILINE)

    @override
    def get_version(self) -> str:
        content = self.file_path.read_text()
        match = self.version_regex.search(content)
        if not match:
            msg = f"Version not found in {self.file_path}"
            raise ValueError(msg)
        return match.group(1)

    @override
    def set_version(self, version: str) -> None:
        content = self.file_path.read_text()
        new_content = self.version_regex.sub(f'__version__ = "{version}"', content)
        _ = self.file_path.write_text(new_content)



"""General utils without dependencies."""

import operator
from collections.abc import Callable, Iterable, Mapping, Sequence
from itertools import pairwise, starmap
from typing import Any, Literal, overload

from kajihs_utils.protocols import (
    SupportsDunderLT,
)


@overload
def get_first[K, V, D](d: Mapping[K, V], /, keys: Iterable[K], default: D = None) -> V | D: ...


@overload
def get_first[K, V, D](
    d: Mapping[K, V],
    /,
    keys: Iterable[K],
    default: Any = None,  # noqa: ANN401 # TODO: Check if it's possible to annotate default with D instead
    *,
    no_default: Literal[True] = True,
) -> V: ...


@overload
def get_first[K, V, D](
    d: Mapping[K, V], /, keys: Iterable[K], default: D = None, *, no_default: bool = True
) -> V | D: ...


def get_first[K, V, D](
    d: Mapping[K, V], /, keys: Iterable[K], default: D = None, *, no_default: bool = False
) -> V | D:
    """
    Return the value of the first key that exists in the mapping.

    Args:
        d: The dictionary to search in.
        keys: The sequence of keys to look for.
        default: The value to return if no key is found.
        no_default: If `True`, raises a `KeyError` if no key is found.

    Returns:
        The value associated with the first found key, or the default value if not found.

    Raises:
        KeyError: If `no_default` is `True` and none of the keys are found.

    Examples:
        >>> d = {"a": 1, "b": 2, "c": 3}
        >>> get_first(d, ["x", "a", "b"])
        1
        >>> get_first(d, ["x", "y"], default=0)
        0
        >>> get_first(d, ["x", "y"], no_default=True)  # Raises: KeyError
        Traceback (most recent call last):
        ...
        KeyError: "None of the keys ['x', 'y'] were found in the mapping."
    """
    for key in keys:
        if key in d:
            return d[key]

    if no_default:
        msg = f"None of the keys {list(keys)} were found in the mapping."
        raise KeyError(msg)

    return default


def is_sorted(values: Iterable[SupportsDunderLT[Any]], /, *, reverse: bool = False) -> bool:
    """
    Determine if the given iterable is sorted in ascending or descending order.

    Args:
        values: An iterable of comparable items supporting the < operator.
        reverse: If False (default), checks for non-decreasing order; if True,
            checks for non-increasing order.

    Returns:
        True if the sequence is sorted according to the given order, False otherwise.

    Examples:
        >>> is_sorted([1, 2, 2, 3])
        True
        >>> is_sorted([3, 2, 1], reverse=True)
        True
        >>> is_sorted([2, 1, 3])
        False
        >>> is_sorted([])
        True
        >>> is_sorted([42])
        True
        >>> # Works with generators as well
        >>> is_sorted(x * x for x in [1, 2, 3, 4])
        True
        >>> # Equal elements are considered sorted
        >>> is_sorted([1, 1, 1])
        True
    """
    op = operator.le if not reverse else operator.ge
    return all(starmap(op, pairwise(values)))


def bisect_predicate[T](
    seq: Sequence[T],
    predicate: Callable[[T], bool],
    lo: int = 0,
    hi: int | None = None,
) -> int:
    """
    Find the first index where predicate flips from True to False using binary search.

    In other words: Find first index where predicate(seq[i]) becomes False.

    The sequence must be partitioned such that all elements where predicate(item) is True
    appear before elements where predicate(item) is False. This is a generalized version
    of bisect_right that works with arbitrary predicates rather than comparison operators.

    Args:
        seq: Partitioned sequence to search. Must have all True-predicate elements
            before any False-predicate elements.
        predicate: Function that returns True for elements that should be considered
            "left" of the insertion point. Typically a condition like lambda x: x < target.
        lo: Lower bound index to start search (inclusive).
        hi: Upper bound index to end search (exclusive). Defaults to len(seq).

    Returns:
        Insertion point index where predicate first fails. This will be:
        - 0 if predicate fails for all elements
        - len(seq) if predicate holds for all elements
        - First index where predicate(seq[i]) == False otherwise

    Examples:
        Find first non-positive number (predicate=lambda x: x <= 0):
        >>> bisect_predicate([-5, -3, 0, 2, 5], lambda x: x <= 0)
        3

        All elements satisfy predicate:
        >>> bisect_predicate([True, True, True], lambda b: b)
        3

        Edge case - empty sequence:
        >>> bisect_predicate([], lambda x: True)
        0

        Custom search range:
        >>> bisect_predicate([1, 3, 5, 7, 9], lambda x: x < 6, lo=1, hi=4)
        3

    Note:
        Similar to bisect.bisect_right but works with arbitrary predicates.
        Requires the array to be properly partitioned - undefined behavior otherwise.
    """
    hi = hi or len(seq)

    while lo < hi:
        mid = (lo + hi) // 2
        if predicate(seq[mid]):
            # print(f"Can complete {mid}")
            lo = mid + 1
        else:
            # print(f"Can't complete {mid}")
            hi = mid
    return lo



"""Utils for matplotlib.pyplot."""

from typing import Any, cast

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from numpy import ndarray

from .arithmetic import (
    almost_factors,
)

type AxesFlatArray = ndarray[tuple[int], Any]


def auto_subplot(
    size: int,
    /,
    ratio: float = 9 / 16,
    *,
    more_cols: bool = False,
    transposed: bool = False,
    return_all_axes: bool = False,
    **subplot_params: Any,
) -> tuple[Figure, AxesFlatArray]:
    """
    Automatically creates a subplot grid with an adequate number of rows and columns.

    Args:
        size: The total number of subplots.
        ratio: The threshold aspect ratio between rows and columns.
        more_cols: Whether there should be columns than rows instead of the
            opposite
        transposed: Whether to transpose the indexes before flattening.
        return_all_axes: Whether to return axis beyond size in the flatten axes.
        **subplot_params: Additional keyword parameters for subplot.

    Returns:
        Tuple containing the figure and the flatten axes.
    """
    # Special case for 2
    large, small = (2, 1) if size == 2 else almost_factors(size, ratio)  # noqa: PLR2004
    rows, cols = (small, large) if more_cols else (large, small)

    fig, axes = plt.subplots(rows, cols, **subplot_params)

    # if isinstance(axes, np.ndarray):
    #     axes = axes.flatten()
    if transposed:
        axes = axes.T
    axes: AxesFlatArray = axes.flatten()

    # Hide the remaining axes if there are more axes than subplots
    for i in range(size, len(axes)):
        axes[i].set_axis_off()

    axes = axes if return_all_axes else cast(AxesFlatArray, axes[:size])  # noqa: TC006

    return fig, axes
